import pandas as pd
import itertools
from pandas.api.types import is_numeric_dtype

def link_dataframes(A: pd.DataFrame, B: pd.DataFrame, ref_col: str, unique_col=None, metric=None) -> pd.DataFrame:
    '''
    Merge two DataFrames A and B according to the reference colum based on minimum metric.

    :param ref_col: Reference Column to merge dataframes. Has to exist in both frames
    :param metric: Metric used to determine matches in ref_col. Default lambda a, b: (a - b).abs()
    :param unique_per_index_col: Is needed for scanner where

    :return: Single merged dataframe consisting of a multiindex from A and B
    '''

    #Check if the Reference column is unique
    if not len(A[ref_col].unique()) == len(A[ref_col]) and not unique_col:
        raise ValueError("Duplicates in ref_col of dataframe A - set unique col")

    try:
        A[ref_col].iloc[0] - B[ref_col].iloc[0]
    except Exception:
        raise ValueError("Reference columns has to be numeric")

    #Generate help DataFrame by merging
    merged_AB =pd.merge(
        A.assign(merge_key=1),
        B.assign(merge_key=1),
        on='merge_key',
        suffixes=('_A','_B')
    )

    #Use absolute Distance as Default Metric
    if not metric:
        metric = lambda a, b: (a - b).abs()

    #Needed because in measurement dataframes from scanner indices and timestamps are not unique
    groupby_keys = [f"{ref_col}_A"]
    if unique_col:
        groupby_keys.append(unique_col)

    #Keep the minimum distance entry for each group
    merged_AB = merged_AB.groupby(by=groupby_keys, axis=0, sort=False).apply(
        lambda grouped: grouped.loc[
            metric(
                grouped[f'{ref_col}_A'], grouped[f'{ref_col}_B']
            ).idxmin()
        ]
    )

    #Remove help columns and set index of merged dataframe to index of A
    merged_AB.drop(columns=['merge_key'], inplace=True)
    merged_AB.reset_index(drop="True", inplace=True)
    merged_AB.set_index(A.index, inplace=True)

    #Generate a Multiindex to seperate columns from dataframes A and B
    merged_AB.columns = pd.MultiIndex.from_tuples((
        itertools.chain(
            zip( len(A.columns)*['A'], A.columns ),
            zip( len(B.columns)*['B'], B.columns )
        )
    ))

    return merged_AB
